/*!
  @file
  @author Shin'ichiro Nakaoka
*/

#include "AISTSimulatorItem.h"
#include "BodyMotionItem.h"
#include "ControllerItem.h"
#include <cnoid/ItemManager>
#include <cnoid/Archive>
#include <cnoid/EigenArchive>
#include <cnoid/World>
#include <cnoid/ConstraintForceSolver>
#include <cnoid/Link>
#include <cnoid/FloatingNumberString>
#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>
#include <fstream>
#include <iomanip>
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;

// for Windows
#undef min
#undef max

namespace {

    const bool TRACE_FUNCTIONS = false;
    const bool ENABLE_DEBUG_OUTPUT = false;
    const double DEFAULT_GRAVITY_ACCELERATION = 9.80665;
    enum DynamicsMode { FORWARD_DYNAMICS = 0, HG_DYNAMICS, KINEMATICS, N_DYNAMICS_MODES };

    class HighGainControllerItem : public ControllerItem
    {
        int currentFrame;
        MultiValueSeqPtr qseqRef;
        int numJoints;
        BodyPtr body;

    public:
        HighGainControllerItem(BodyItem* bodyItem, BodyMotionItem* bodyMotionItem) {
            qseqRef = bodyMotionItem->jointPosSeq();
            setName(str(fmt(_("HighGain Controller with %1%")) % bodyMotionItem->name()));
        }

        virtual bool initialize(Body* body, double worldTimeStep) {
            this->body = body;
            currentFrame = 0;
            numJoints = std::min(body->numJoints(), qseqRef->numParts());
            if(qseqRef->numFrames() == 0){
                putMessage(_("Reference motion is empty()."));
                return false;
            }
            if(fabs(qseqRef->frameRate() - (1.0 / worldTimeStep)) > 1.0e-6){
                putMessage(_("The frame rate of the reference motion is different from the world frame rate."));
                return false;
            }
            output();
            return true;
        }

        virtual double timeStep() {
            return qseqRef->getTimeStep();
        }
        
        virtual void input() { }

        virtual bool control() {
            bool doContinue = true;
            const int n = qseqRef->numFrames();
            if(++currentFrame >= n){
                currentFrame = n -1;
                return false;
            }
            return true;
        }
        
        virtual void output() {
            MultiValueSeq::Frame qRef = qseqRef->frame(currentFrame);
            double frameRate = qseqRef->frameRate();
            for(int i=0; i < numJoints; ++i){
                Link* joint = body->joint(i);
                double qNext = qRef[i];
                double dqNext = (qNext - joint->q) * frameRate; 
                joint->ddq = (dqNext - joint->dq) * frameRate;
                joint->dq = dqNext;
                joint->q = qNext;
            }
        }
        
        virtual void finalize() { }

    };
}


namespace cnoid {
  
    class AISTSimulatorItemImpl
    {
    public:
        AISTSimulatorItem* self;

        World<ConstraintForceSolver> world;
        
        Selection dynamicsMode;
        Vector3 gravity;
        double staticFriction;
        double slipFriction;
        double cullingThresh;
        FloatingNumberString errorCriterion;
        int maxNumIterations;
        FloatingNumberString contactCorrectionDepth;
        FloatingNumberString contactCorrectionVelocityRatio;
        double epsilon;
        bool is2Dmode;

        typedef std::map<Body*, int> BodyIndexMap;
        BodyIndexMap bodyIndexMap;

        AISTSimulatorItemImpl(AISTSimulatorItem* self);
        AISTSimulatorItemImpl(AISTSimulatorItem* self, const AISTSimulatorItemImpl& org);
        ControllerItem* createBodyMotionController(BodyItem* bodyItem, BodyMotionItem* bodyMotionItem);
        bool initializeSimulation(std::vector<SimulationBody*>& simBodies);
        void addBody(SimulationBodyPtr simBody);
        bool stepSimulation(std::vector<SimulationBody*>& activeSimBodies);
        bool onFrictionPropertyChanged(double mu, int type);
        bool onCullingThreshChanged(double thresh);
        bool on2DmodeChanged(bool on);

        void doPutProperties(PutPropertyFunction& putProperty);
        bool store(Archive& archive);
        bool restore(const Archive& archive);

        // for debug
        ofstream os;
    };
}


void AISTSimulatorItem::initialize(ExtensionManager* ext)
{
    ext->itemManager().registerClass<AISTSimulatorItem>(N_("AISTSimulatorItem"));
    ext->itemManager().addCreationPanel<AISTSimulatorItem>();
}


AISTSimulatorItem::AISTSimulatorItem()
{
    impl = new AISTSimulatorItemImpl(this);
}


AISTSimulatorItemImpl::AISTSimulatorItemImpl(AISTSimulatorItem* self)
    : self(self),
      dynamicsMode(N_DYNAMICS_MODES, CNOID_GETTEXT_DOMAIN_NAME)
{
    dynamicsMode.setSymbol(FORWARD_DYNAMICS,  N_("Forward dynamics"));
    dynamicsMode.setSymbol(HG_DYNAMICS,       N_("High-gain dynamics"));
    dynamicsMode.setSymbol(KINEMATICS,        N_("Kinematics"));

    gravity << 0.0, 0.0, -DEFAULT_GRAVITY_ACCELERATION;
    staticFriction = 1.0;
    slipFriction = 1.0;
    cullingThresh = 0.02;
    errorCriterion = ConstraintForceSolver::defaultGaussSeidelErrorCriterion();
    maxNumIterations = ConstraintForceSolver::defaultGaussSeidelMaxNumIterations();
    contactCorrectionDepth = ConstraintForceSolver::defaultContactCorrectionDepth();
    contactCorrectionVelocityRatio = ConstraintForceSolver::defaultContactCorrectionVelocityRatio();
    
    epsilon = 0.0;
    is2Dmode = false;
}


AISTSimulatorItem::AISTSimulatorItem(const AISTSimulatorItem& org)
    : SimulatorItem(org),
      impl(new AISTSimulatorItemImpl(this, *org.impl))
{

}


AISTSimulatorItemImpl::AISTSimulatorItemImpl(AISTSimulatorItem* self, const AISTSimulatorItemImpl& org)
    : self(self),
      dynamicsMode(org.dynamicsMode)
{
    gravity = org.gravity;
    staticFriction = org.staticFriction;
    slipFriction = org.slipFriction;
    cullingThresh = org.cullingThresh;
    errorCriterion = org.errorCriterion;
    maxNumIterations = org.maxNumIterations;
    contactCorrectionDepth = org.contactCorrectionDepth;
    contactCorrectionVelocityRatio = org.contactCorrectionVelocityRatio;
    epsilon = org.epsilon;
    is2Dmode = org.is2Dmode;
}


AISTSimulatorItem::~AISTSimulatorItem()
{
    delete impl;
}


ItemPtr AISTSimulatorItem::doDuplicate() const
{
    return new AISTSimulatorItem(*this);
}


SimulationBody* AISTSimulatorItem::createSimulationBody()
{
    return new SimulationBody();
}


ControllerItem* AISTSimulatorItem::createBodyMotionController(BodyItem* bodyItem, BodyMotionItem* bodyMotionItem)
{
    return new HighGainControllerItem(bodyItem, bodyMotionItem);
}


bool AISTSimulatorItem::initializeSimulation(std::vector<SimulationBody*>& simBodies)
{
    return impl->initializeSimulation(simBodies);
}


bool AISTSimulatorItemImpl::initializeSimulation(std::vector<SimulationBody*>& simBodies)
{
    if(ENABLE_DEBUG_OUTPUT){
        static int ntest = 0;
        os.open((string("test-log-") + lexical_cast<string>(ntest++) + ".log").c_str());
        os << setprecision(30);
    }
    
    world.setRungeKuttaMethod();
    world.setGravityAcceleration(gravity);
    world.enableSensors(true);
    world.setTimeStep(self->worldTimeStep());
    world.setCurrentTime(0.0);

    world.constraintForceSolver.setGaussSeidelErrorCriterion(errorCriterion.value());
    world.constraintForceSolver.setGaussSeidelMaxNumIterations(maxNumIterations);
    world.constraintForceSolver.setContactDepthCorrection(
        contactCorrectionDepth.value(), contactCorrectionVelocityRatio.value());

    world.clearBodies();
    bodyIndexMap.clear();
    for(size_t i=0; i < simBodies.size(); ++i){
        addBody(simBodies[i]);
    }

    world.constraintForceSolver.clearCollisionCheckLinkPairs();
    std::vector<std::pair<Link*,Link*> > pairs;
    self->getColdetLinkPairs(pairs);
    for(size_t i=0; i < pairs.size(); ++i){
        Link* link1 = pairs[i].first;
        Link* link2 = pairs[i].second;
        int body1Index = bodyIndexMap[link1->body];
        int body2Index = bodyIndexMap[link2->body];
        world.constraintForceSolver.addCollisionCheckLinkPair
            (body1Index, link1, body2Index, link2, staticFriction, slipFriction, cullingThresh, epsilon);
    }

    if(is2Dmode){
        world.constraintForceSolver.set2Dmode(true);
    }

    world.initialize();

    if(ENABLE_DEBUG_OUTPUT){
        os << "initializeSimulation\n" ;
        for(size_t i=0; i < simBodies.size(); ++i){
            simBodies[i]->body()->putInformation(os);
        }
    }

    return true;
}


void AISTSimulatorItemImpl::addBody(SimulationBodyPtr simBody)
{
    Body* body = simBody->body();

    Link* rootLink = body->rootLink();
    rootLink->v.setZero();
    rootLink->dv.setZero();
    rootLink->w.setZero();
    rootLink->dw.setZero();
    rootLink->vo.setZero();
    rootLink->dvo.setZero();

    bool isHighGainMode = dynamicsMode.is(HG_DYNAMICS);
    if(dynamic_cast<HighGainControllerItem*>(simBody->controller())){
        isHighGainMode = true;
    }

    // Using clone internal models may be better to avoid a bug caused by
    // accessing the same object from multi-threads.
    //for(int i=0; i < body->numLinks(); ++i){
    //    body->link(i)->coldetModel->cloneInternalModel();
    //}

    for(int i=0; i < body->numJoints(); ++i){
        Link* joint = body->joint(i);
        joint->isHighGainMode = isHighGainMode;
        joint->u = 0.0;
        joint->dq = 0.0;
        joint->ddq = 0.0;
    }
    
    body->clearExternalForces();
    body->calcForwardKinematics(true, true);

    bodyIndexMap[body] = world.addBody(body);
}


bool AISTSimulatorItem::stepSimulation(std::vector<SimulationBody*>& activeSimBodies)
{
    return impl->stepSimulation(activeSimBodies);
}


bool AISTSimulatorItemImpl::stepSimulation(std::vector<SimulationBody*>& activeSimBodies)
{
    world.constraintForceSolver.clearExternalForces();

    if(dynamicsMode.is(KINEMATICS)){
        for(size_t i=0; i < activeSimBodies.size(); ++i){
            activeSimBodies[i]->body()->calcForwardKinematics();
        }
    } else {
        world.calcNextState();
    }

    if(ENABLE_DEBUG_OUTPUT){
        os << "stepSimulation: frame = " << self->currentFrame() << "\n";
        for(size_t i=0; i < activeSimBodies.size(); ++i){
            activeSimBodies[i]->body()->putInformation(os);
        }
    }

    return true;
}


void AISTSimulatorItem::finalizeSimulation()
{
    if(ENABLE_DEBUG_OUTPUT){
        impl->os.close();
    }
}


bool AISTSimulatorItemImpl::onFrictionPropertyChanged(double mu, int type)
{
    if(mu > 0.0){
        if(type == 0){
            staticFriction = mu;
        } else if(type == 1){
            slipFriction = mu;
        }
        return true;
    }
    return false;
}


bool AISTSimulatorItemImpl::onCullingThreshChanged(double thresh)
{
    if(thresh >= 0.0){
        cullingThresh = thresh;
        return true;
    }
    return false;
}


bool AISTSimulatorItemImpl::on2DmodeChanged(bool on)
{
    is2Dmode = on;
    return true;
}


void AISTSimulatorItem::doPutProperties(PutPropertyFunction& putProperty)
{
    SimulatorItem::doPutProperties(putProperty);
    impl->doPutProperties(putProperty);
}


void AISTSimulatorItemImpl::doPutProperties(PutPropertyFunction& putProperty)
{
    putProperty(_("Dynamics mode"), dynamicsMode,
                bind((bool(Selection::*)(int))&Selection::select, &dynamicsMode, _1));
    putProperty(_("Gravity"), str(gravity), bind(toVector3, _1, ref(gravity)));
    putProperty(_("Static friction"), staticFriction,
                (bind(&AISTSimulatorItemImpl::onFrictionPropertyChanged, this, _1, 0)));
    putProperty(_("Slip friction"), slipFriction,
                (bind(&AISTSimulatorItemImpl::onFrictionPropertyChanged, this, _1, 1)));
    putProperty(_("Culling thresh"), cullingThresh,
                (bind(&AISTSimulatorItemImpl::onCullingThreshChanged, this, _1)));
    putProperty(_("Error criterion"), errorCriterion,
                bind(&FloatingNumberString::setPositiveValue, ref(errorCriterion), _1));
    putProperty.min(1.0)(_("Max iterations"), maxNumIterations, changeProperty(maxNumIterations));
    putProperty(_("CC depth"), contactCorrectionDepth,
                bind(&FloatingNumberString::setNonNegativeValue, ref(contactCorrectionDepth), _1));
    putProperty(_("CC v-ratio"), contactCorrectionVelocityRatio,
                bind(&FloatingNumberString::setPositiveValue, ref(contactCorrectionVelocityRatio), _1));
    putProperty(_("2D mode"), is2Dmode,
                (bind(&AISTSimulatorItemImpl::on2DmodeChanged, this, _1)));
}


bool AISTSimulatorItem::store(Archive& archive)
{
    SimulatorItem::store(archive);
    return impl->store(archive);
}


bool AISTSimulatorItemImpl::store(Archive& archive)
{
    archive.write("dynamicsMode", dynamicsMode.selectedSymbol());
    writeMatrix(archive, "gravity", gravity);
    archive.write("staticFriction", staticFriction);
    archive.write("slipFriction", slipFriction);
    archive.write("cullingThresh", cullingThresh);
    archive.write("errorCriterion", errorCriterion);
    archive.write("maxNumIterations", maxNumIterations);
    archive.write("contactCorrectionDepth", contactCorrectionDepth);
    archive.write("contactCorrectionVelocityRatio", contactCorrectionVelocityRatio);
    archive.write("2Dmode", is2Dmode);
    return true;
}


bool AISTSimulatorItem::restore(const Archive& archive)
{
    SimulatorItem::restore(archive);
    return impl->restore(archive);
}


bool AISTSimulatorItemImpl::restore(const Archive& archive)
{
    string symbol;
    if(archive.read("dynamicsMode", symbol)){
        dynamicsMode.select(symbol);
    }
    read(archive, "gravity", gravity);
    archive.read("staticFriction", staticFriction);
    archive.read("slipFriction", slipFriction);
    archive.read("cullingThresh", cullingThresh);
    errorCriterion = archive.get("errorCriterion", errorCriterion.string());
    archive.read("maxNumIterations", maxNumIterations);
    contactCorrectionDepth = archive.get("contactCorrectionDepth", contactCorrectionDepth.string());
    contactCorrectionVelocityRatio = archive.get("contactCorrectionVelocityRatio", contactCorrectionVelocityRatio.string());
    archive.read("2Dmode", is2Dmode);
    return true;
}
